import random
import string
from typing import Callable, List, Tuple, Union

OTRawListType = List[Union[int, str, dict]]


def make_random_doc(amount: int) -> str:
    if amount < 10:
        amount = 10

    return ''.join(random.choices(string.ascii_letters, k=amount))


def make_random_ots(
    normalize: Callable[[OTRawListType], OTRawListType],
    doc: str,
    n: int,
    ids_weights: Tuple[float, float, float] = (0.4, 0.4, 0.2)
) -> OTRawListType:
    offset = 0
    ot_raw_list: List[Union[int, str, dict]] = []
    for _ in range(n):
        if len(doc) - offset == 0:
            action, = random.choices(['i', 'd'], ids_weights[:2], k=1)
        else:
            action, = random.choices(['i', 'd', 's'], ids_weights, k=1)

        if action == 'i':
            amount = random.randint(1, max(1, min(len(doc) - offset, len(doc) // n)))
            ot_raw_list.append(''.join(random.choices(string.ascii_letters, k=amount)))

        elif action == 'd':
            amount = random.randint(1, max(1, min(len(doc) - offset, len(doc) // n)))
            ot_raw_list.append({'d': doc[offset: offset + amount]})
            offset += amount

        elif action == 's':
            amount = random.randint(1, max(1, min(len(doc) - offset, len(doc) // n)))
            ot_raw_list.append(amount)
            offset += amount

    return normalize(ot_raw_list)



'''Python implementation of Operational Transform(OT) text type

Original:
https://github.com/ottypes/docs
https://github.com/ottypes/text/blob/master/lib/text.js
'''


OP_SKIP = 0
OP_DELETE = 1
OP_INSERT = 2


def _resolve_op(op):
    if isinstance(op, int):
        return (OP_SKIP, op)
    elif isinstance(op, dict):
        return (OP_DELETE, op.get('d', ''))
    elif isinstance(op, str):
        return (OP_INSERT, op)
    return (None, None)


def _make_op(op_type, op_arg):
    if op_type == OP_SKIP:
        return op_arg
    elif op_type == OP_DELETE:
        return {'d': op_arg}
    elif op_type == OP_INSERT:
        return op_arg
    return None


def check(ops):
    if not isinstance(ops, list):
        return False

    last_op_type = None
    op_type = None

    for op in ops:
        resolved_op = _resolve_op(op)
        if resolved_op is None:
            # unknown op
            return False

        op_type, op_arg = resolved_op

        if op_type == OP_SKIP:
            if not isinstance(op_arg, int) or op_arg <= 0:
                # print('no skip', op)
                return False
        elif op_type == OP_DELETE:
            if not isinstance(op_arg, str) or not op_arg:
                # print('wrong delete', op)
                return False
        elif op_type == OP_INSERT:
            if not isinstance(op_arg, str) or not op_arg:
                # print('wrong insert', op)
                return False

        if last_op_type == op_type:
            # un-optimized ops
            # print('dup op')
            return False

        last_op_type = op_type

    return True


def _make_iter_ops(ops):
    for op in ops:
        resolved_op = _resolve_op(op)
        if resolved_op is None:
            break
        yield resolved_op


def make_appender(ops):
    def appender(op):
        if op is None:
            return ops

        op_type, op_arg = _resolve_op(op)

        if ops:
            last_op_type, last_op_arg = _resolve_op(ops[-1])
        else:
            last_op_type, last_op_arg = None, None

        if op_arg == 0 or op_arg == '':
            return ops

        if last_op_type == op_type:
            merged_op_arg = last_op_arg + op_arg
            ops[-1] = _make_op(op_type, merged_op_arg)
        else:
            ops.append(op)

        return ops

    return appender


def make_taker(ops):
    context = [0, 0]

    def inner_taker(n, indivisable=None):
        idx, offset = context

        if idx == len(ops):
            if n == -1:
                return None
            return n

        ret = None
        op_type, op_arg = _resolve_op(ops[idx])

        if op_type == OP_SKIP:
            if n == -1 or op_arg - offset <= n:
                ret = _make_op(OP_SKIP, op_arg - offset)
                idx += 1
                offset = 0
            else:
                ret = _make_op(OP_SKIP, n)
                offset += n

        elif op_type == OP_DELETE:
            if n == -1 or indivisable == OP_DELETE or len(op_arg) - offset <= n:
                ret = _make_op(OP_DELETE, op_arg[offset:])
                offset = 0
                idx += 1
            else:
                ret = _make_op(OP_DELETE, op_arg[offset:offset + n])
                offset += n

        elif op_type == OP_INSERT:
            if n == -1 or indivisable == OP_INSERT or len(op_arg) - offset <= n:
                ret = _make_op(OP_INSERT, op_arg[offset:])
                offset = 0
                idx += 1
            else:
                ret = _make_op(OP_INSERT, op_arg[offset:offset + n])
                offset += n

        context[0] = idx
        context[1] = offset
        return ret

    def inner_peeker():
        idx, _ = context
        if 0 <= idx < len(ops):
            return ops[idx]
        else:
            return None

    return inner_taker, inner_peeker


def trim(normalized_ops):
    '''Trim ops in place

    Discrade trailing OP_SKIPs in ops.
    `normalized_ops` must be normalized.
    '''
    if normalized_ops and _resolve_op(normalized_ops[-1])[0] == OP_SKIP:
        normalized_ops.pop()
    return normalized_ops


def normalize(ops):
    '''Normalize ops

    Merge consecutive operations and trim the result.
    '''
    new_ops = []
    appender = make_appender(new_ops)

    for op in ops:
        appender(op)

    return trim(new_ops)


def apply(doc, ops):
    '''Apply ops to doc
    '''

    if not isinstance(doc, str):
        raise ValueError('doc must be string')

    if not check(ops):
        raise ValueError('invalid ops')

    new_doc = []

    for op_type, op_arg in _make_iter_ops(ops):
        if op_type == OP_SKIP:
            if op_arg > len(doc):
                raise ValueError('the op is too long')

            new_doc.append(doc[:op_arg])
            doc = doc[op_arg:]

        elif op_type == OP_DELETE:
            if doc[:len(op_arg)] != op_arg:
                # race condition : deleting text was changed by others
                raise ValueError('inconsistent delete', doc[:len(op_arg)], op_arg)

            doc = doc[len(op_arg):]

        elif op_type == OP_INSERT:
            new_doc.append(op_arg)

    new_doc.append(doc)

    return ''.join(new_doc)


def inverse_apply(doc, ops):
    '''Inversely apply ops to doc
    '''

    if not isinstance(doc, str):
        raise ValueError('doc must be string')

    if not check(ops):
        raise ValueError('invalid ops')

    last_pos = 0

    for op_type, op_arg in _make_iter_ops(ops):
        if op_type == OP_SKIP:
            last_pos += op_arg

        elif op_type == OP_DELETE:
            pass

        elif op_type == OP_INSERT:
            last_pos += len(op_arg)

    old_doc = [doc[last_pos:]]
    doc = doc[:last_pos]

    for op_type, op_arg in _make_iter_ops(reversed(ops)):
        if op_type == OP_SKIP:
            old_doc.append(doc[-op_arg:])
            doc = doc[:-op_arg]

        elif op_type == OP_DELETE:
            old_doc.append(op_arg)

        elif op_type == OP_INSERT:
            if doc[-len(op_arg):] != op_arg:
                raise ValueError('inconsistent state', doc[-len(op_arg):], op_arg)

            doc = doc[:-len(op_arg)]

    old_doc.append(doc)

    return ''.join(reversed(old_doc))



import random
import time
from typing import TYPE_CHECKING, Any, List

from ottype import core
from tests import utils

NUM_ITERATION = 100_000

CORE_IMPL: List[Any] = [('python', core)]

try:
    from ottype import core_boost
    CORE_IMPL.append(('cython', core_boost))
except ImportError:
    pass

try:
    from extra import old_ottype
    CORE_IMPL.insert(0, ('baseline', old_ottype))
except ImportError:
    pass

random.seed(457700)


def benchmark_apply() -> None:
    print('Target: apply')

    prev_record = dict()

    for core_imple_name, core_impl in CORE_IMPL:
        print(f'=== {core_imple_name} ===')

        if TYPE_CHECKING:
            from ottype import core as core_impl  # type:ignore

        for doc_length in [100, 1_000, 10_000]:
            doc = utils.make_random_doc(doc_length)

            for ot_length in [5, 10, 20, 50, 100]:
                exp_ident = (doc_length, ot_length)

                ots = utils.make_random_ots(core_impl.normalize, doc, ot_length)

                st_time = time.perf_counter_ns()
                for _ in range(NUM_ITERATION):
                    core_impl.apply(doc, ots)
                ed_time = time.perf_counter_ns()

                perf = (ed_time - st_time) / 1_000 / NUM_ITERATION
                if exp_ident not in prev_record:
                    prev_record[exp_ident] = perf

                print('len(doc) : %5d, len(OTs) : %3d, Performance : %6.2lf ms/loop ( %5.2lfx )' % (
                    doc_length, ot_length, perf, 1 / (perf / prev_record[exp_ident])
                ))


def benchmark_inverse_apply() -> None:
    print('Target: inverse_apply')

    prev_record = dict()

    for core_imple_name, core_impl in CORE_IMPL:
        print(f'=== {core_imple_name} ===')

        if TYPE_CHECKING:
            from ottype import core as core_impl  # type:ignore

        for doc_length in [100, 1_000, 10_000]:
            doc = utils.make_random_doc(doc_length)

            for ot_length in [5, 10, 20, 50, 100]:
                exp_ident = (doc_length, ot_length)

                ots = utils.make_random_ots(core_impl.normalize, doc, ot_length)
                new_doc = core_impl.apply(doc, ots)

                st_time = time.perf_counter_ns()
                for _ in range(NUM_ITERATION):
                    core_impl.inverse_apply(new_doc, ots)
                ed_time = time.perf_counter_ns()

                perf = (ed_time - st_time) / 1_000 / NUM_ITERATION
                if exp_ident not in prev_record:
                    prev_record[exp_ident] = perf

                print('len(doc) : %5d, len(OTs) : %3d, Performance : %6.2lf ms/loop ( %5.2lfx )' % (
                    doc_length, ot_length, perf, 1 / (perf / prev_record[exp_ident])
                ))


benchmark_apply()
print()
benchmark_inverse_apply()



# cython: language_level=3, boundscheck=False
from cpython cimport *

cdef enum OTTypeAction:
    nop = 0, skip = 1, insert = 2, delete = 3


def  _resolve_ot(object ot_raw):
    if isinstance(ot_raw, int):
        if ot_raw <= 0:
            raise ValueError('invalid OT-Skip')
        return OTTypeAction.skip, ot_raw
    elif isinstance(ot_raw, str):
        if ot_raw == '':
            raise ValueError('invalid OT-Insert')
        return OTTypeAction.insert, ot_raw
    elif isinstance(ot_raw, dict):
        s = ot_raw.get('d')
        if not isinstance(s, str) or s == '':
            raise ValueError('invalid OT-Delete')
        return OTTypeAction.delete, s
    elif isinstance(ot_raw, tuple):
        if len(ot_raw) == 2:
            ot_action, ot_arg = ot_raw
            if ot_action == 1:
                assert isinstance(ot_arg, int)
                if ot_arg <= 0:
                    raise ValueError('invalid OT-Skip')
                return ot_raw
            elif ot_action == 2:
                assert isinstance(ot_arg, str)
                if ot_arg == '':
                    raise ValueError('invalid OT-Insert')
                return ot_raw
            elif ot_action == 3:
                assert isinstance(ot_arg, str)
                if ot_arg == '':
                    raise ValueError('invalid OT-Delete')
                return ot_raw

    raise ValueError('unexpected OT structure')


def _to_ot_raw_list(list ots):
    cdef:
        Py_ssize_t ots_length, i
        OTTypeAction ot_action
        object ot_arg
        object ot_raw

    ots_length = PyList_Size(ots)
    ot_raw_list = PyList_New(ots_length)

    for i in range(ots_length):
        ot_action, ot_arg = <tuple>PyList_GET_ITEM(ots, i)
        if ot_action == OTTypeAction.skip:
            ot_raw = ot_arg
        elif ot_action == OTTypeAction.insert:
            ot_raw = ot_arg
        elif ot_action == OTTypeAction.delete:
            ot_raw = {'d': ot_arg}

        Py_INCREF(ot_raw)
        PyList_SET_ITEM(ot_raw_list, i, ot_raw)

    return ot_raw_list


cdef class _Appender:
    cdef list ots

    def __init__(self, list ots):
        self.ots = ots

    def append(self, tuple ot):
        cdef:
            OTTypeAction last_ot_action
            object last_ot_arg

            OTTypeAction ot_action
            object ot_arg

        if ot is None:
            return
        
        if not self.ots:
            self.ots.append(ot)
            return

        ot_action, ot_arg = ot
        last_ot_action, last_ot_arg = <tuple>self.ots[-1]

        if last_ot_action == OTTypeAction.skip and ot_action == OTTypeAction.skip:
            self.ots[-1] = (OTTypeAction.skip, <int>last_ot_arg + <int>ot_arg)
        elif last_ot_action == OTTypeAction.insert and ot_action == OTTypeAction.insert:
            self.ots[-1] = (OTTypeAction.insert, <str>last_ot_arg + <str>ot_arg)
        elif last_ot_action == OTTypeAction.delete and ot_action == OTTypeAction.delete:
            self.ots[-1] = (OTTypeAction.delete, <str>last_ot_arg + <str>ot_arg)
        else:
            self.ots.append(ot)


cdef class _Taker:
    cdef:
        object ot_raw_list
        int _idx
        int _offset
    
    def __init__(self, object ot_raw_list):
        self.ot_raw_list = ot_raw_list

        self._idx = 0
        self._offset = 0

    def take(self, int n, str indivisable = None):
        cdef:
            tuple ret_ot
            OTTypeAction ot_action
            object ot_arg
            int ot_arg_as_int
            str ot_arg_as_str

        if self._idx == len(self.ot_raw_list):
            if n == -1:
                return None
            return (OTTypeAction.skip, n)

        ot_action, ot_arg = _resolve_ot(self.ot_raw_list[self._idx])
        ret_ot = None

        if ot_action == OTTypeAction.skip:
            ot_arg_as_int = <int>ot_arg

            if n == -1 or ot_arg_as_int - self._offset <= n:
                ret_ot = (OTTypeAction.skip, ot_arg_as_int - self._offset)
                self._idx += 1
                self._offset = 0
            else:
                ret_ot = (OTTypeAction.skip, n)
                self._offset += n

        elif ot_action == OTTypeAction.insert:
            ot_arg_as_str = <str>ot_arg

            if n == -1 or indivisable == 'i' or len(ot_arg_as_str) - self._offset <= n:
                ret_ot = (OTTypeAction.insert, ot_arg_as_str[self._offset:])
                self._idx += 1
                self._offset = 0
            else:
                ret_ot = (OTTypeAction.insert, ot_arg_as_str[self._offset:self._offset + n])
                self._offset += n

        elif ot_action == OTTypeAction.delete:
            ot_arg_as_str = <str>ot_arg

            if n == -1 or indivisable == 'd' or len(ot_arg_as_str) - self._offset <= n:
                ret_ot = (OTTypeAction.delete, ot_arg_as_str[self._offset:])
                self._idx += 1
                self._offset = 0
            else:
                ret_ot = (OTTypeAction.delete, ot_arg_as_str[self._offset:self._offset + n])
                self._offset += n

        return ret_ot

    def peak_action(self):
        if 0 <= self._idx < len(self.ot_raw_list):
            return _resolve_ot(self.ot_raw_list[self._idx])[0]
        return OTTypeAction.nop


def _trim(list ots not None):
    if ots and ots[-1][0] == OTTypeAction.skip:
        ots.pop()


def check(object ot_raw_list not None, *, bool check_unoptimized not None = True):
    cdef:
        OTTypeAction last_ot_action
        OTTypeAction ot_action

    if not isinstance(ot_raw_list, (list, tuple)):
        raise TypeError('`ot_raw_list` must be a list or tuple')

    last_ot_action = OTTypeAction.nop
    try:
        for ot_raw in ot_raw_list:
            ot_action = _resolve_ot(ot_raw)[0]

            if check_unoptimized and last_ot_action == ot_action:
                return False

            last_ot_action = ot_action

    except (ValueError, TypeError):
        return False

    if check_unoptimized and last_ot_action == OTTypeAction.skip:
        return False

    return True


def apply(str doc not None, object ot_raw_list not None, *, bool check_unoptimized not None = True):
    cdef:
        list new_doc
        int pos

        OTTypeAction ot_action
        object ot_arg
        int ot_arg_as_int
        str ot_arg_as_str

    if not isinstance(ot_raw_list, (list, tuple)):
        raise TypeError('`ot_raw_list` must be a list or tuple')

    if not check(ot_raw_list, check_unoptimized=check_unoptimized):
        raise ValueError('invalid OTs')

    new_doc = []
    pos = 0

    for ot_raw in ot_raw_list:
        ot_action, ot_arg = _resolve_ot(ot_raw)

        if ot_action == OTTypeAction.skip:
            ot_arg_as_int = <int>ot_arg

            if ot_arg_as_int > len(doc) - pos:
                raise ValueError('skip exceeds doc length')

            new_doc.append(doc[pos:pos + ot_arg_as_int])
            pos += ot_arg_as_int

        elif ot_action == OTTypeAction.insert:
            ot_arg_as_str = <str>ot_arg

            new_doc.append(ot_arg_as_str)

        elif ot_action == OTTypeAction.delete:
            ot_arg_as_str = <str>ot_arg

            if doc[pos:pos + len(ot_arg_as_str)] != ot_arg_as_str:
                raise ValueError(
                    'inconsistent delete (doc, OT.arg)',
                    doc[pos:pos + len(ot_arg_as_str)],
                    ot_arg_as_str,
                )
            pos += len(ot_arg_as_str)

    new_doc.append(doc[pos:])

    return ''.join(new_doc)


def inverse_apply(str doc not None, object ot_raw_list not None, *, bool check_unoptimized not None = True):
    cdef:
        list ot_list

        int last_pos
        list old_doc

        OTTypeAction ot_action
        object ot_arg
        int ot_arg_as_int
        str ot_arg_as_str

    if not isinstance(ot_raw_list, (list, tuple)):
        raise TypeError('`ot_raw_list` must be a list or tuple')

    if not check(ot_raw_list, check_unoptimized=check_unoptimized):
        raise ValueError('invalid OTs')

    last_pos = 0

    for ot_raw in ot_raw_list:
        ot_action, ot_arg = _resolve_ot(ot_raw)

        if ot_action == OTTypeAction.skip:
            last_pos += <int>ot_arg

        elif ot_action == OTTypeAction.insert:
            last_pos += len(<str>ot_arg)

        elif ot_action == OTTypeAction.delete:
            pass

    if last_pos > len(doc):
        raise ValueError('skip exceeds doc length')

    old_doc = [doc[last_pos:]]

    for ot_raw in reversed(ot_raw_list):
        ot_action, ot_arg = _resolve_ot(ot_raw)

        if ot_action == OTTypeAction.skip:
            ot_arg_as_int = <int>ot_arg

            old_doc.append(doc[last_pos - ot_arg_as_int:last_pos])
            last_pos -= ot_arg_as_int

        elif ot_action == OTTypeAction.insert:
            ot_arg_as_str = <str>ot_arg

            if doc[last_pos - len(ot_arg_as_str):last_pos] != ot_arg_as_str:
                raise ValueError(
                    'inconsistent delete (doc, OT.arg)',
                    doc[last_pos - len(ot_arg_as_str):last_pos],
                    ot_arg_as_str,
                )
            last_pos -= len(ot_arg_as_str)

        elif ot_action == OTTypeAction.delete:
            ot_arg_as_str = <str>ot_arg

            old_doc.append(ot_arg_as_str)

    old_doc.append(doc[:last_pos])

    return ''.join(reversed(old_doc))


def normalize(object ot_raw_list not None):
    cdef:
        list new_ots

    if not isinstance(ot_raw_list, (list, tuple)):
        raise TypeError('`ot_raw_list` must be a list or tuple')

    if not check(ot_raw_list, check_unoptimized=False):
        raise ValueError('invalid OTs')

    new_ots = []
    appender = _Appender(new_ots)
    for ot_raw in ot_raw_list:
        appender.append(_resolve_ot(ot_raw))

    _trim(new_ots)

    return _to_ot_raw_list(new_ots)


def transform(object ot_raw_list_1 not None, object ot_raw_list_2 not None, str side not None):
    cdef:
        list new_ots
        _Appender appender
        _Taker taker

        OTTypeAction ot_action
        object ot_arg
        int n

        tuple chunk_ot
        OTTypeAction chunk_ot_action
        object chunk_ot_arg

    if not isinstance(ot_raw_list_1, (list, tuple)):
        raise TypeError('`ot_raw_list_1` must be a list or tuple')

    if not isinstance(ot_raw_list_2, (list, tuple)):
        raise TypeError('`ot_raw_list_2` must be a list or tuple')

    if not check(ot_raw_list_1) or not check(ot_raw_list_2):
        raise ValueError('invalid OTs')

    if side not in ['left', 'right']:
        raise ValueError('invalid side')

    new_ots = []
    appender = _Appender(new_ots)
    taker = _Taker(ot_raw_list_1)

    for ot_raw in ot_raw_list_2:
        ot_action, ot_arg = _resolve_ot(ot_raw)

        if ot_action == OTTypeAction.skip:
            n = <int>ot_arg

            while 0 < n:
                chunk_ot = taker.take(n, 'i')
                appender.append(chunk_ot)

                if chunk_ot is None:
                    break

                chunk_ot_action, chunk_ot_arg = chunk_ot

                if chunk_ot_action == OTTypeAction.skip:
                    n -= <int>chunk_ot_arg
                elif chunk_ot_action == OTTypeAction.insert:
                    pass
                elif chunk_ot_action == OTTypeAction.delete:
                    n -= len(<str>chunk_ot_arg)

        elif ot_action == OTTypeAction.insert:
            n = len(<str>ot_arg)

            if (
                side == 'left'
                and taker.peak_action() == OTTypeAction.insert
            ):
                appender.append(taker.take(-1))

            appender.append((OTTypeAction.skip, n))

        elif ot_action == OTTypeAction.delete:
            n = len(<str>ot_arg)

            while 0 < n:
                chunk_ot = taker.take(n, 'i')
                chunk_ot_action, chunk_ot_arg = chunk_ot

                if chunk_ot_action == OTTypeAction.skip:
                    n -= <int>chunk_ot_arg
                elif chunk_ot_action == OTTypeAction.insert:
                    appender.append(chunk_ot)
                elif chunk_ot_action == OTTypeAction.delete:
                    n -= len(<str>chunk_ot_arg)

    while True:
        chunk_ot = taker.take(-1)
        if chunk_ot is None:
            break
        appender.append(chunk_ot)

    _trim(new_ots)

    return _to_ot_raw_list(new_ots)


def compose(object ot_raw_list_1 not None, object ot_raw_list_2 not None):
    cdef:
        list new_ots
        _Appender appender
        _Taker taker

        tuple ot
        OTTypeAction ot_action
        object ot_arg
        str ot_arg_as_str

        int n
        int offset

        tuple chunk_ot
        OTTypeAction chunk_ot_action
        object chunk_ot_arg
        int chunk_ot_arg_as_int
        str chunk_ot_arg_as_str

    if not isinstance(ot_raw_list_1, (list, tuple)):
        raise TypeError('`ot_raw_list_1` must be a list or tuple')

    if not isinstance(ot_raw_list_2, (list, tuple)):
        raise TypeError('`ot_raw_list_2` must be a list or tuple')

    if not check(ot_raw_list_1) or not check(ot_raw_list_2):
        raise ValueError('invalid OTs')

    new_ots = []
    appender = _Appender(new_ots)
    taker = _Taker(ot_raw_list_1)

    for ot_raw in ot_raw_list_2:
        ot_action, ot_arg = ot = _resolve_ot(ot_raw)

        if ot_action == OTTypeAction.skip:
            n = <int>ot_arg

            while 0 < n:
                chunk_ot = taker.take(n, 'd')
                appender.append(chunk_ot)

                chunk_ot_action, chunk_ot_arg = chunk_ot

                if chunk_ot_action == OTTypeAction.skip:
                    n -= <int>chunk_ot_arg
                elif chunk_ot_action == OTTypeAction.insert:
                    n -= len(<str>chunk_ot_arg)
                elif chunk_ot_action == OTTypeAction.delete:
                    pass

        elif ot_action == OTTypeAction.insert:
            appender.append(ot)

        elif ot_action == OTTypeAction.delete:
            ot_arg_as_str = <str>ot_arg

            offset = 0
            n = len(ot_arg)

            while 0 < n:
                chunk_ot = taker.take(n, 'd')
                chunk_ot_action, chunk_ot_arg = chunk_ot

                if chunk_ot_action == OTTypeAction.skip:
                    chunk_ot_arg_as_int = <int>chunk_ot_arg

                    appender.append((OTTypeAction.delete, ot_arg_as_str[offset:offset + chunk_ot_arg_as_int]))
                    offset += chunk_ot_arg_as_int
                    n -= chunk_ot_arg_as_int

                elif chunk_ot_action == OTTypeAction.insert:
                    chunk_ot_arg_as_str = <str>chunk_ot_arg

                    if chunk_ot_arg_as_str != ot_arg_as_str[offset:offset + len(chunk_ot_arg_as_str)]:
                        raise ValueError(
                            'inconsistent delete in the seconds OTs (doc, OT.arg)',
                            chunk_ot_arg_as_str,
                            ot_arg_as_str[offset:offset + len(chunk_ot_arg_as_str)],
                        )
                    offset += len(chunk_ot_arg_as_str)
                    n -= len(chunk_ot_arg_as_str)

                elif chunk_ot_action == OTTypeAction.delete:
                    appender.append(chunk_ot)

    while True:
        chunk_ot = taker.take(-1)
        if chunk_ot is None:
            break
        appender.append(chunk_ot)

    _trim(new_ots)

    return _to_ot_raw_list(new_ots)



from __future__ import annotations

from typing import TYPE_CHECKING, List

import pytest

from ottype import core

from . import utils

FUZZ_TEST_COUNT = 1_000
FUZZ_TEST_INIT_DOC_LENGTH = 100
FUZZ_TEST_OTS_LENGTH = 20

CORE_IMPL = [core]
try:
    from ottype import core_boost
    CORE_IMPL.append(core_boost)
except ImportError:
    pass


@pytest.fixture(params=CORE_IMPL)
def core_impl(request):  # type:ignore
    return request.param


@pytest.fixture(params=[list, tuple])
def input_cls(request):  # type:ignore
    return request.param


def OTSkip(arg: int) -> core._OTType:
    return (core._OTTypeActionSkip, arg)


def OTInsert(arg: str) -> core._OTType:
    return (core._OTTypeActionInsert, arg)


def OTDelete(arg: str) -> core._OTType:
    return (core._OTTypeActionDelete, arg)


def test__resolve_ot(core_impl) -> None:  # type: ignore
    if TYPE_CHECKING:
        core_impl = core

    _resolve_ot = core_impl._resolve_ot

    # OT-Skip

    assert _resolve_ot(3) == OTSkip(3)
    with pytest.raises(ValueError):
        _resolve_ot(-3)

    # OT-Insert

    assert _resolve_ot('asdf') == OTInsert('asdf')
    with pytest.raises(ValueError):
        _resolve_ot('')

    # OT-Delete

    assert _resolve_ot({'d': 'asdf'}) == OTDelete('asdf')

    with pytest.raises(ValueError):
        _resolve_ot({'a': 'asdf'})
    with pytest.raises(ValueError):
        _resolve_ot({'d': ''})
    with pytest.raises(ValueError):
        _resolve_ot({'d': 4})

    # Raw
    assert _resolve_ot((1, 3)) == OTSkip(3)
    assert _resolve_ot((2, 'asdf')) == OTInsert('asdf')
    assert _resolve_ot((3, 'asdf')) == OTDelete('asdf')

    with pytest.raises(ValueError):
        _resolve_ot((0, 1234))

    with pytest.raises(ValueError):
        _resolve_ot((4, 'asdf'))

    # ETC

    with pytest.raises(ValueError):
        _resolve_ot(3.141592)  # type: ignore


def test__Appender(core_impl) -> None:  # type: ignore
    if TYPE_CHECKING:
        core_impl = core

    _Appender = core_impl._Appender

    ots_1: List[core._OTType] = []
    appender_1 = _Appender(ots_1)

    appender_1.append(None)
    assert ots_1 == []

    appender_1.append(OTSkip(4))
    assert ots_1 == [OTSkip(4)]

    appender_1.append(OTSkip(3))
    assert ots_1 == [OTSkip(7)]

    appender_1.append(OTInsert('as'))
    assert ots_1 == [OTSkip(7), OTInsert('as')]

    appender_1.append(OTInsert('df'))
    assert ots_1 == [OTSkip(7), OTInsert('asdf')]

    appender_1.append(OTDelete('qw'))
    assert ots_1 == [OTSkip(7), OTInsert('asdf'), OTDelete('qw')]

    appender_1.append(OTDelete('er'))
    assert ots_1 == [OTSkip(7), OTInsert('asdf'), OTDelete('qwer')]


def test__Taker(core_impl) -> None:  # type: ignore
    if TYPE_CHECKING:
        core_impl = core

    _Taker = core_impl._Taker

    ots_1: List[core._OTRawInputType] = [3, 4]
    taker_1 = _Taker(ots_1)
    assert taker_1.take(1) == OTSkip(1)
    assert taker_1.take(4) == OTSkip(2)
    assert taker_1.take(-1) == OTSkip(4)
    assert taker_1.take(-1) is None
    assert taker_1.take(5) == OTSkip(5)

    ots_2: List[core._OTRawInputType] = ['asdf', 'qwer', 'zxcv']
    taker_2 = _Taker(ots_2)
    assert taker_2.take(1) == OTInsert('a')
    assert taker_2.take(1) == OTInsert('s')
    assert taker_2.take(5) == OTInsert('df')
    assert taker_2.take(1) == OTInsert('q')
    assert taker_2.take(1, 'd') == OTInsert('w')
    assert taker_2.take(1, 'i') == OTInsert('er')
    assert taker_2.take(-1) == OTInsert('zxcv')

    ots_3: List[core._OTRawInputType] = [{'d': 'asdf'}, {'d': 'qwer'}, {'d': 'zxcv'}]
    taker_3 = _Taker(ots_3)
    assert taker_3.take(1) == OTDelete('a')
    assert taker_3.take(1) == OTDelete('s')
    assert taker_3.take(5) == OTDelete('df')
    assert taker_3.take(1) == OTDelete('q')
    assert taker_3.take(1, 'i') == OTDelete('w')
    assert taker_3.take(1, 'd') == OTDelete('er')
    assert taker_3.take(-1) == OTDelete('zxcv')

    ots_4: List[core._OTRawInputType] = [3, 'asdf']
    taker_4 = _Taker(ots_4)
    assert taker_4.peak_action() == core._OTTypeActionSkip
    taker_4.take(1)
    assert taker_4.peak_action() == core._OTTypeActionSkip
    taker_4.take(5)
    assert taker_4.peak_action() == core._OTTypeActionInsert
    taker_4.take(3)
    assert taker_4.peak_action() == core._OTTypeActionInsert
    taker_4.take(1)
    assert taker_4.peak_action() == core._OTTypeActionNop


def test__trim(core_impl) -> None:  # type: ignore
    if TYPE_CHECKING:
        core_impl = core

    _trim = core_impl._trim

    ots_1: List[core._OTType] = []
    _trim(ots_1)
    assert ots_1 == []

    ots_2: List[core._OTType] = [OTSkip(4)]
    _trim(ots_2)
    assert ots_2 == []

    ots_3: List[core._OTType] = [OTInsert('asdf'), OTSkip(4)]
    _trim(ots_3)
    assert ots_3 == [OTInsert('asdf')]

    ots_4: List[core._OTType] = [OTSkip(4), OTInsert('asdf')]
    _trim(ots_4)
    assert ots_4 == [OTSkip(4), OTInsert('asdf')]


def test_check(core_impl, input_cls) -> None:  # type:ignore
    if TYPE_CHECKING:
        core_impl = core
        input_cls = list

    check = core_impl.check

    with pytest.raises(TypeError):
        check(4)  # type: ignore

    assert check(input_cls([3, 'asdf', {'d': 'qwer'}]))
    assert not check(input_cls([3, object()]))  # type: ignore
    assert not check(input_cls([3, 4]))
    assert not check(input_cls([3]))


def test_apply(core_impl, input_cls) -> None:  # type:ignore
    if TYPE_CHECKING:
        core_impl = core
        input_cls = list

    apply = core_impl.apply

    with pytest.raises(TypeError):
        apply(None, [])  # type: ignore

    with pytest.raises(TypeError):
        apply('', 12345)  # type: ignore

    with pytest.raises(ValueError):
        apply('', input_cls([3, 4]))

    assert apply('abcde', input_cls([2, 'qq', {'d': 'c'}, 1, 'w'])) == 'abqqdwe'

    with pytest.raises(ValueError):
        apply('aa', input_cls([3, 'x']))

    with pytest.raises(ValueError):
        apply('aa', input_cls([{'d': 'b'}]))


def test_apply_fuzz(core_impl, input_cls) -> None:  # type:ignore
    if TYPE_CHECKING:
        core_impl = core

    apply = core_impl.apply
    normalize = core_impl.normalize

    for _ in range(FUZZ_TEST_COUNT):
        doc = utils.make_random_doc(FUZZ_TEST_INIT_DOC_LENGTH)
        random_ot_raw_list = input_cls(utils.make_random_ots(normalize, doc, FUZZ_TEST_OTS_LENGTH))

        apply(doc, random_ot_raw_list)


def test_inverse_apply(core_impl, input_cls) -> None:  # type:ignore
    if TYPE_CHECKING:
        core_impl = core
        input_cls = list

    inverse_apply = core_impl.inverse_apply

    with pytest.raises(TypeError):
        inverse_apply(None, [])  # type: ignore

    with pytest.raises(TypeError):
        inverse_apply('', 12345)  # type: ignore

    assert inverse_apply('abqqdwe', input_cls([2, 'qq', {'d': 'c'}, 1, 'w'])) == 'abcde'

    with pytest.raises(ValueError):
        inverse_apply('', input_cls([3, 4]))

    with pytest.raises(ValueError):
        inverse_apply('aa', input_cls([3, 'x']))

    with pytest.raises(ValueError):
        inverse_apply('aa', input_cls(['b']))


def test_inverse_apply_fuzz(core_impl, input_cls) -> None:  # type:ignore
    if TYPE_CHECKING:
        core_impl = core

    apply = core_impl.apply
    inverse_apply = core_impl.inverse_apply
    normalize = core_impl.normalize

    for _ in range(FUZZ_TEST_COUNT):
        doc = utils.make_random_doc(FUZZ_TEST_INIT_DOC_LENGTH)
        random_ot_raw_list = input_cls(utils.make_random_ots(normalize, doc, FUZZ_TEST_OTS_LENGTH))

        new_doc = apply(doc, random_ot_raw_list)
        assert doc == inverse_apply(new_doc, random_ot_raw_list)


def test_normalize(core_impl, input_cls) -> None:  # type:ignore
    if TYPE_CHECKING:
        core_impl = core
        input_cls = list

    normalize = core_impl.normalize

    with pytest.raises(TypeError):
        normalize(12345)  # type:ignore

    with pytest.raises(ValueError):
        normalize(input_cls([-123]))

    assert normalize(input_cls([
        3, 4,
        'as', 'df',
        {'d': 'qw'}, {'d': 'er'},
        5
    ])) == [7, 'asdf', {'d': 'qwer'}]


def test_transform_fuzz(core_impl, input_cls) -> None:  # type:ignore
    if TYPE_CHECKING:
        core_impl = core

    apply = core_impl.apply
    normalize = core_impl.normalize
    transform = core_impl.transform

    with pytest.raises(TypeError):
        transform(1234, [3], 'left')

    with pytest.raises(TypeError):
        transform([4], 1234, 'left')

    with pytest.raises(TypeError):
        transform([3], [4], None)

    with pytest.raises(ValueError):
        transform([3, 4], [], 'left')

    with pytest.raises(ValueError):
        transform([], [3, 4], 'left')

    with pytest.raises(ValueError):
        transform([], [], 'good')  # type: ignore

    for _ in range(FUZZ_TEST_COUNT):
        doc = utils.make_random_doc(FUZZ_TEST_INIT_DOC_LENGTH)
        random_ot_raw_list_1 = input_cls(utils.make_random_ots(normalize, doc, FUZZ_TEST_OTS_LENGTH))
        random_ot_raw_list_2 = input_cls(utils.make_random_ots(normalize, doc, FUZZ_TEST_OTS_LENGTH))

        left_first_doc = apply(
            apply(doc, random_ot_raw_list_1),
            transform(random_ot_raw_list_2, random_ot_raw_list_1, 'left')
        )

        right_first_doc = apply(
            apply(doc, random_ot_raw_list_2),
            transform(random_ot_raw_list_1, random_ot_raw_list_2, 'right')
        )

        assert left_first_doc == right_first_doc


def test_compose_fuzz(core_impl, input_cls) -> None:  # type:ignore
    if TYPE_CHECKING:
        core_impl = core

    apply = core_impl.apply
    compose = core_impl.compose
    normalize = core_impl.normalize

    with pytest.raises(TypeError):
        compose(1234, [4])

    with pytest.raises(TypeError):
        compose([3], 1234)

    with pytest.raises(ValueError):
        compose([3, 4], [])

    with pytest.raises(ValueError):
        compose([], [3, 4])

    with pytest.raises(ValueError):
        compose(['asdf'], [{'d': 'qw'}])

    for _ in range(FUZZ_TEST_COUNT):
        doc_1 = utils.make_random_doc(FUZZ_TEST_INIT_DOC_LENGTH)
        random_ot_raw_list_1 = input_cls(utils.make_random_ots(normalize, doc_1, FUZZ_TEST_OTS_LENGTH))

        doc_2 = apply(doc_1, random_ot_raw_list_1)
        random_ot_raw_list_2 = input_cls(utils.make_random_ots(normalize, doc_2, FUZZ_TEST_OTS_LENGTH))

        doc_3 = apply(doc_2, random_ot_raw_list_2)

        doc_3_composed = apply(
            doc_1,
            compose(
                random_ot_raw_list_1,
                random_ot_raw_list_2
            )
        )

        assert doc_3 == doc_3_composed


def test_complex_fuzz(core_impl, input_cls) -> None:  # type:ignore
    if TYPE_CHECKING:
        core_impl = core

    apply = core_impl.apply
    compose = core_impl.compose
    normalize = core_impl.normalize
    transform = core_impl.transform

    for _ in range(FUZZ_TEST_COUNT):
        # Situation: A local editor receives `server_ot_raw_list_2` from a server
        #            while the editor holds `local_ot_raw_list_1` and `local_ot_raw_list_1a` in a buffer
        #            that is not sended to the server. How does the editor transform the buffer?

        doc = utils.make_random_doc(FUZZ_TEST_INIT_DOC_LENGTH)
        local_ot_raw_list_1 = input_cls(utils.make_random_ots(normalize, doc, FUZZ_TEST_OTS_LENGTH))
        local_ot_raw_list_1a = input_cls(
            utils.make_random_ots(normalize, apply(doc, local_ot_raw_list_1), FUZZ_TEST_OTS_LENGTH)
        )

        server_ot_raw_list_2 = input_cls(utils.make_random_ots(normalize, doc, FUZZ_TEST_OTS_LENGTH))

        left_first_doc = apply(
            apply(
                apply(doc, local_ot_raw_list_1),
                local_ot_raw_list_1a
            ),
            transform(
                server_ot_raw_list_2,
                compose(
                    local_ot_raw_list_1,
                    local_ot_raw_list_1a
                ),
                'left'
            )
        )

        right_first_doc = apply(
            apply(doc, server_ot_raw_list_2),
            transform(
                compose(
                    local_ot_raw_list_1,
                    local_ot_raw_list_1a
                ),
                server_ot_raw_list_2,
                'right'
            )
        )

        assert left_first_doc == right_first_doc



import os
import re
import sys
from os import path

from setuptools import Extension, setup

wdir = path.abspath(path.dirname(__file__))

with open(path.join(wdir, 'ottype', '__init__.py'), encoding='utf-8') as f:
    try:
        version = re.findall(r"^__version__ = '([^']+)'\r?$", f.read(), re.M)[0]
    except IndexError:
        raise RuntimeError('Unable to determine version.')

with open(path.join(wdir, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


NO_EXTENSIONS = bool(os.environ.get('OTTYPE_NO_EXTENSIONS'))
if sys.implementation.name != 'cpython':
    NO_EXTENSIONS = True

try:
    from Cython.Build import cythonize
except ImportError:
    NO_EXTENSIONS = True

dev_install_requires = [
    'autopep8==1.6.0',
    'bandit==1.7.4',
    'Cython==0.29.28',
    'flake8==4.0.1',
    'flake8-bugbear==22.1.11',
    'flake8-isort==4.1.1',
    'flake8-quotes==3.3.1',
    'mypy==0.931',
    'pytest==7.0.1',
    'pytest-cov==3.0.0',
]

if not NO_EXTENSIONS:
    ext_modules = cythonize([
        Extension('ottype.core_boost', ['ottype/core_boost.pyx'])
    ])
else:
    ext_modules = []

setup(
    name='python-ottype',
    version=version,
    description='A python implementation of Operational Transformation.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    keywords='ottype',

    author='Jungkook Park',
    author_email='pjknkda@gmail.com',
    url='https://github.com/pjknkda/python-ottype',
    license='MIT',

    packages=['ottype'],
    package_data={'ottype': ['py.typed']},
    zip_safe=False,

    python_requires='>=3.7, <3.11',

    extras_require={'dev': dev_install_requires},

    ext_modules=ext_modules,

    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)



import os

from .core import apply as _apply_py
from .core import check as _check_py
from .core import compose as _compose_py
from .core import inverse_apply as _inverse_apply_py
from .core import normalize as _normalize_py
from .core import transform as _transform_py

__version__ = '22.3.1'

NO_EXTENSIONS = bool(os.environ.get('OTTYPE_NO_EXTENSIONS'))


apply = _apply_py
check = _check_py
compose = _compose_py
inverse_apply = _inverse_apply_py
normalize = _normalize_py
transform = _transform_py

try:
    if not NO_EXTENSIONS:
        from .core_boost import apply as _apply_c
        from .core_boost import check as _check_c
        from .core_boost import compose as _compose_c
        from .core_boost import inverse_apply as _inverse_apply_c
        from .core_boost import normalize as _normalize_c
        from .core_boost import transform as _transform_c

        apply = _apply_c
        check = _check_c
        compose = _compose_c
        inverse_apply = _inverse_apply_c
        normalize = _normalize_c
        transform = _transform_c

except ImportError:  # pragma: no cover
    pass



from __future__ import annotations

from typing import TYPE_CHECKING, Dict, List, NewType, Optional, Tuple, Union

if TYPE_CHECKING:
    from typing_extensions import Literal


_OTTypeAction = NewType('_OTTypeAction', int)

_OTTypeActionNop = _OTTypeAction(0)
_OTTypeActionSkip = _OTTypeAction(1)
_OTTypeActionInsert = _OTTypeAction(2)
_OTTypeActionDelete = _OTTypeAction(3)

_OTType = Tuple[_OTTypeAction, Union[int, str]]

_OTRawInputType = Union[int, str, Dict[str, str], Tuple[int, Union[int, str]]]
_OTRawInputList = Union[List[_OTRawInputType], Tuple[_OTRawInputType]]

_OTRawOutputType = Union[int, str, Dict[str, str]]


def _resolve_ot(ot_raw: _OTRawInputType) -> _OTType:
    if isinstance(ot_raw, int):
        if ot_raw <= 0:
            raise ValueError('invalid OT-Skip')
        return (_OTTypeActionSkip, ot_raw)
    elif isinstance(ot_raw, str):
        if ot_raw == '':
            raise ValueError('invalid OT-Insert')
        return (_OTTypeActionInsert, ot_raw)
    elif isinstance(ot_raw, dict):
        s = ot_raw.get('d')
        if not isinstance(s, str) or s == '':
            raise ValueError('invalid OT-Delete')
        return (_OTTypeActionDelete, s)
    elif isinstance(ot_raw, tuple):
        if len(ot_raw) == 2:
            ot_action, ot_arg = ot_raw
            if ot_action == 1:
                assert isinstance(ot_arg, int)
                if ot_arg <= 0:
                    raise ValueError('invalid OT-Skip')
                return ot_raw  # type: ignore
            elif ot_action == 2:
                assert isinstance(ot_arg, str)
                if ot_arg == '':
                    raise ValueError('invalid OT-Insert')
                return ot_raw  # type: ignore
            elif ot_action == 3:
                assert isinstance(ot_arg, str)
                if ot_arg == '':
                    raise ValueError('invalid OT-Delete')
                return ot_raw  # type: ignore

    raise ValueError('unexpected OT structure')


def _to_ot_raw_list(ots: List[_OTType]) -> List[_OTRawOutputType]:
    ot_raw_list = []
    for ot_action, ot_arg in ots:
        ot_raw: _OTRawOutputType

        if ot_action == _OTTypeActionSkip:
            assert isinstance(ot_arg, int)
            ot_raw = ot_arg

        elif ot_action == _OTTypeActionInsert:
            assert isinstance(ot_arg, str)
            ot_raw = ot_arg

        elif ot_action == _OTTypeActionDelete:
            assert isinstance(ot_arg, str)
            ot_raw = {'d': ot_arg}

        else:
            raise ValueError('unexpected OT')

        ot_raw_list.append(ot_raw)

    return ot_raw_list


class _Appender:
    def __init__(self, ots: List[_OTType]) -> None:
        self.ots = ots

    def append(self, ot: Optional[_OTType]) -> None:
        if ot is None:
            return

        if not self.ots:
            self.ots.append(ot)
            return

        last_ot_action, last_ot_arg = self.ots[-1]
        ot_action, ot_arg = ot

        if last_ot_action == _OTTypeActionSkip and ot_action == _OTTypeActionSkip:
            assert isinstance(last_ot_arg, int)
            assert isinstance(ot_arg, int)
            self.ots[-1] = (_OTTypeActionSkip, last_ot_arg + ot_arg)
        elif last_ot_action == _OTTypeActionInsert and ot_action == _OTTypeActionInsert:
            assert isinstance(last_ot_arg, str)
            assert isinstance(ot_arg, str)
            self.ots[-1] = (_OTTypeActionInsert, last_ot_arg + ot_arg)
        elif last_ot_action == _OTTypeActionDelete and ot_action == _OTTypeActionDelete:
            assert isinstance(last_ot_arg, str)
            assert isinstance(ot_arg, str)
            self.ots[-1] = (_OTTypeActionDelete, last_ot_arg + ot_arg)
        else:
            self.ots.append(ot)


class _Taker:
    def __init__(self, ot_raw_list: _OTRawInputList) -> None:
        self.ot_raw_list = ot_raw_list

        self._idx = 0
        self._offset = 0

    def take(self, n: int, indivisable: Optional[Literal['d', 'i']] = None) -> Optional[_OTType]:
        if self._idx == len(self.ot_raw_list):
            if n == -1:
                return None
            return (_OTTypeActionSkip, n)

        ot_action, ot_arg = _resolve_ot(self.ot_raw_list[self._idx])
        ret_ot: Optional[_OTType] = None

        if ot_action == _OTTypeActionSkip:
            assert isinstance(ot_arg, int)
            if n == -1 or ot_arg - self._offset <= n:
                ret_ot = (_OTTypeActionSkip, ot_arg - self._offset)
                self._idx += 1
                self._offset = 0
            else:
                ret_ot = (_OTTypeActionSkip, n)
                self._offset += n

        elif ot_action == _OTTypeActionInsert:
            assert isinstance(ot_arg, str)
            if n == -1 or indivisable == 'i' or len(ot_arg) - self._offset <= n:
                ret_ot = (_OTTypeActionInsert, ot_arg[self._offset:])
                self._idx += 1
                self._offset = 0
            else:
                ret_ot = (_OTTypeActionInsert, ot_arg[self._offset:self._offset + n])
                self._offset += n

        elif ot_action == _OTTypeActionDelete:
            assert isinstance(ot_arg, str)
            if n == -1 or indivisable == 'd' or len(ot_arg) - self._offset <= n:
                ret_ot = (_OTTypeActionDelete, ot_arg[self._offset:])
                self._idx += 1
                self._offset = 0
            else:
                ret_ot = (_OTTypeActionDelete, ot_arg[self._offset:self._offset + n])
                self._offset += n

        return ret_ot

    def peak_action(self) -> _OTTypeAction:
        if 0 <= self._idx < len(self.ot_raw_list):
            return _resolve_ot(self.ot_raw_list[self._idx])[0]
        return _OTTypeActionNop


def _trim(ots: List[_OTType]) -> None:
    '''Trim ots in place

    Discrade trailing OP_SKIPs in ots.
    `ots` must be normalized.
    '''
    if ots and ots[-1][0] == _OTTypeActionSkip:
        ots.pop()


def check(ot_raw_list: _OTRawInputList, *, check_unoptimized: bool = True) -> bool:
    if not isinstance(ot_raw_list, (list, tuple)):
        raise TypeError('`ot_raw_list` must be a list or tuple')

    last_ot_action = _OTTypeActionNop
    try:
        for ot_raw in ot_raw_list:
            ot_action = _resolve_ot(ot_raw)[0]

            if check_unoptimized and last_ot_action == ot_action:
                # un-optimized ots
                return False

            last_ot_action = ot_action

    except (ValueError, TypeError):
        return False

    if check_unoptimized and last_ot_action == _OTTypeActionSkip:
        return False

    return True


def apply(
    doc: str,
    ot_raw_list: _OTRawInputList,
    *,
    check_unoptimized: bool = True,
) -> str:
    '''Apply ots to doc
    '''

    if not isinstance(doc, str):
        raise TypeError('`doc` must be string')

    if not isinstance(ot_raw_list, (list, tuple)):
        raise TypeError('`ot_raw_list` must be a list or tuple')

    if not check(ot_raw_list, check_unoptimized=check_unoptimized):
        raise ValueError('invalid OTs')

    new_doc = []
    pos = 0

    for ot_raw in ot_raw_list:
        ot_action, ot_arg = _resolve_ot(ot_raw)

        if ot_action == _OTTypeActionSkip:
            assert isinstance(ot_arg, int)

            if ot_arg > len(doc) - pos:
                raise ValueError('skip exceeds doc length')

            new_doc.append(doc[pos:pos + ot_arg])
            pos += ot_arg

        elif ot_action == _OTTypeActionInsert:
            assert isinstance(ot_arg, str)

            new_doc.append(ot_arg)

        elif ot_action == _OTTypeActionDelete:
            assert isinstance(ot_arg, str)

            if doc[pos:pos + len(ot_arg)] != ot_arg:
                raise ValueError('inconsistent delete (doc, OT.arg)', doc[pos:pos + len(ot_arg)], ot_arg)
            pos += len(ot_arg)

    new_doc.append(doc[pos:])

    return ''.join(new_doc)


def inverse_apply(
    doc: str,
    ot_raw_list: _OTRawInputList,
    *,
    check_unoptimized: bool = True,
) -> str:
    '''Inversely apply ots to doc
    '''

    if not isinstance(doc, str):
        raise TypeError('`doc` must be string')

    if not isinstance(ot_raw_list, (list, tuple)):
        raise TypeError('`ot_raw_list` must be a list or tuple')

    if not check(ot_raw_list, check_unoptimized=check_unoptimized):
        raise ValueError('invalid OTs')

    last_pos = 0
    for ot_raw in ot_raw_list:
        ot_action, ot_arg = _resolve_ot(ot_raw)

        if ot_action == _OTTypeActionSkip:
            assert isinstance(ot_arg, int)
            last_pos += ot_arg

        elif ot_action == _OTTypeActionInsert:
            assert isinstance(ot_arg, str)
            last_pos += len(ot_arg)

        elif ot_action == _OTTypeActionDelete:
            pass

    if last_pos > len(doc):
        raise ValueError('skip exceeds doc length')

    old_doc = [doc[last_pos:]]

    for ot_raw in reversed(ot_raw_list):
        ot_action, ot_arg = _resolve_ot(ot_raw)

        if ot_action == _OTTypeActionSkip:
            assert isinstance(ot_arg, int)

            old_doc.append(doc[last_pos - ot_arg:last_pos])
            last_pos -= ot_arg

        elif ot_action == _OTTypeActionInsert:
            assert isinstance(ot_arg, str)

            if doc[last_pos - len(ot_arg):last_pos] != ot_arg:
                raise ValueError('inconsistent delete (doc, OT.arg)', doc[last_pos - len(ot_arg):last_pos], ot_arg)
            last_pos -= len(ot_arg)

        elif ot_action == _OTTypeActionDelete:
            assert isinstance(ot_arg, str)

            old_doc.append(ot_arg)

    old_doc.append(doc[:last_pos])

    return ''.join(reversed(old_doc))


def normalize(ot_raw_list: _OTRawInputList) -> List[_OTRawOutputType]:
    '''Normalize ots

    Merge consecutive operations and trim the result.
    '''

    if not isinstance(ot_raw_list, (list, tuple)):
        raise TypeError('`ot_raw_list` must be a list or tuple')

    if not check(ot_raw_list, check_unoptimized=False):
        raise ValueError('invalid OTs')

    new_ots: List[_OTType] = []
    appender = _Appender(new_ots)
    for ot_raw in ot_raw_list:
        appender.append(_resolve_ot(ot_raw))

    _trim(new_ots)

    return _to_ot_raw_list(new_ots)


def transform(
    ot_raw_list_1: _OTRawInputList,
    ot_raw_list_2: _OTRawInputList,
    side: Literal['left', 'right'],
) -> List[_OTRawOutputType]:
    '''Transform `ot_raw_list_1` by `ot_raw_list_2`

    Transform `ot_raw_list_1` to have same meaning when `ot_raw_list_2` is applied
    to the doc before `ot_raw_list_1`.

    `side` is required to break ties, for example, if we assume
    `ot_raw_list_1 = ['a']` and `ot_raw_list_2 = ['b']`, the result can be either
    'ab' or 'ba' depending on the side.

    - `transform(['a'], ['b'], 'left') = [1, "a"]` (left <- right)
    - `transform(['a'], ['b'], 'right') = ["a"]` (left -> right)

    The result of transform satisfies that,
    .. code::
        apply(apply(doc, local_ots), transform(server_ots, local_ots, 'left'))
            == apply(apply(doc, server_ots), transform(local_ots, server_ots, 'right'))
    '''

    if not isinstance(ot_raw_list_1, (list, tuple)):
        raise TypeError('`ot_raw_list_1` must be a list or tuple')

    if not isinstance(ot_raw_list_2, (list, tuple)):
        raise TypeError('`ot_raw_list_2` must be a list or tuple')

    if not isinstance(side, str):
        raise TypeError('`side` must be str')

    if not check(ot_raw_list_1) or not check(ot_raw_list_2):
        raise ValueError('invalid OTs')

    if side not in ['left', 'right']:
        raise ValueError('invalid side')

    new_ots: List[_OTType] = []
    appender = _Appender(new_ots)
    taker = _Taker(ot_raw_list_1)

    for ot_raw in ot_raw_list_2:
        ot_action, ot_arg = _resolve_ot(ot_raw)

        if ot_action == _OTTypeActionSkip:
            assert isinstance(ot_arg, int)

            n = ot_arg
            while 0 < n:
                chunk_ot = taker.take(n, 'i')
                appender.append(chunk_ot)

                if chunk_ot is None:
                    break  # pragma: no cover

                chunk_ot_action, chunk_ot_arg = chunk_ot

                if chunk_ot_action == _OTTypeActionSkip:
                    assert isinstance(chunk_ot_arg, int)
                    n -= chunk_ot_arg
                elif chunk_ot_action == _OTTypeActionInsert:
                    pass
                elif chunk_ot_action == _OTTypeActionDelete:
                    assert isinstance(chunk_ot_arg, str)
                    n -= len(chunk_ot_arg)

        elif ot_action == _OTTypeActionInsert:
            assert isinstance(ot_arg, str)

            n = len(ot_arg)

            if (
                side == 'left'
                and taker.peak_action() == _OTTypeActionInsert
            ):
                appender.append(taker.take(-1))

            appender.append((_OTTypeActionSkip, n))

        elif ot_action == _OTTypeActionDelete:
            assert isinstance(ot_arg, str)

            n = len(ot_arg)
            while 0 < n:
                chunk_ot = taker.take(n, 'i')

                if chunk_ot is None:
                    break  # pragma: no cover

                chunk_ot_action, chunk_ot_arg = chunk_ot

                if chunk_ot_action == _OTTypeActionSkip:
                    assert isinstance(chunk_ot_arg, int)
                    n -= chunk_ot_arg
                elif chunk_ot_action == _OTTypeActionInsert:
                    appender.append(chunk_ot)
                elif chunk_ot_action == _OTTypeActionDelete:
                    assert isinstance(chunk_ot_arg, str)
                    n -= len(chunk_ot_arg)

    while True:
        chunk_ot = taker.take(-1)
        if chunk_ot is None:
            break
        appender.append(chunk_ot)

    _trim(new_ots)

    return _to_ot_raw_list(new_ots)


def compose(
    ot_raw_list_1: _OTRawInputList,
    ot_raw_list_2: _OTRawInputList,
) -> List[_OTRawOutputType]:
    '''Compose `ot_raw_list_1` and `ot_raw_list_2`

    The result of compose satisfies
    .. code::
        apply(apply(doc, ot_raw_list_1), ot_raw_list_2) == apply(doc, compose(ot_raw_list_1, ot_raw_list_2))
    '''

    if not isinstance(ot_raw_list_1, (list, tuple)):
        raise TypeError('`ot_raw_list_1` must be a list or tuple')

    if not isinstance(ot_raw_list_2, (list, tuple)):
        raise TypeError('`ot_raw_list_2` must be a list or tuple')

    if not check(ot_raw_list_1) or not check(ot_raw_list_2):
        raise ValueError('invalid OTs')

    new_ots: List[_OTType] = []
    appender = _Appender(new_ots)
    taker = _Taker(ot_raw_list_1)

    for ot_raw in ot_raw_list_2:
        ot_action, ot_arg = ot = _resolve_ot(ot_raw)

        if ot_action == _OTTypeActionSkip:
            assert isinstance(ot_arg, int)

            n = ot_arg
            while 0 < n:
                chunk_ot = taker.take(n, 'd')
                appender.append(chunk_ot)

                if chunk_ot is None:
                    break  # pragma: no cover

                chunk_ot_action, chunk_ot_arg = chunk_ot

                if chunk_ot_action == _OTTypeActionSkip:
                    assert isinstance(chunk_ot_arg, int)
                    n -= chunk_ot_arg
                elif chunk_ot_action == _OTTypeActionInsert:
                    assert isinstance(chunk_ot_arg, str)
                    n -= len(chunk_ot_arg)
                elif chunk_ot_action == _OTTypeActionDelete:
                    pass

        elif ot_action == _OTTypeActionInsert:
            appender.append(ot)

        elif ot_action == _OTTypeActionDelete:
            assert isinstance(ot_arg, str)

            offset = 0
            n = len(ot_arg)

            while 0 < n:
                chunk_ot = taker.take(n, 'd')

                if chunk_ot is None:
                    break  # pragma: no cover

                chunk_ot_action, chunk_ot_arg = chunk_ot

                if chunk_ot_action == _OTTypeActionSkip:
                    assert isinstance(chunk_ot_arg, int)

                    appender.append((_OTTypeActionDelete, ot_arg[offset:offset + chunk_ot_arg]))
                    offset += chunk_ot_arg
                    n -= chunk_ot_arg

                elif chunk_ot_action == _OTTypeActionInsert:
                    assert isinstance(chunk_ot_arg, str)

                    if chunk_ot_arg != ot_arg[offset:offset + len(chunk_ot_arg)]:
                        raise ValueError(
                            'inconsistent delete in the seconds OTs (doc, OT.arg)',
                            chunk_ot_arg,
                            ot_arg[offset:offset + len(chunk_ot_arg)],
                        )
                    offset += len(chunk_ot_arg)
                    n -= len(chunk_ot_arg)

                elif chunk_ot_action == _OTTypeActionDelete:
                    appender.append(chunk_ot)

    while True:
        chunk_ot = taker.take(-1)
        if chunk_ot is None:
            break
        appender.append(chunk_ot)

    _trim(new_ots)

    return _to_ot_raw_list(new_ots)
